@@ -10,16 +10,16 @@ mod app;
 mod build;
 mod container_context;
 mod container_port_mapping;
+mod log;
+mod macros;
 mod pack;
 mod util;
 
-pub use crate::container_context::{ContainerContext, ContainerExecResult};
-
-use crate::pack::PackBuildCommand;
-use bollard::container::{Config, CreateContainerOptions, StartContainerOptions};
+pub use crate::container_context::{ContainerContext, PrepareContainerContext};
+use crate::pack::{PackBuildCommand, PullPolicy};
 use bollard::image::RemoveImageOptions;
 use bollard::Docker;
-
+use std::collections::HashMap;
 use std::env;
 use std::env::VarError;
 use std::path::{Path, PathBuf};
@@ -34,7 +34,7 @@ use std::process::{Command, Stdio};
 ///
 /// # Example
 /// ```no_run
-/// use libcnb_test::{IntegrationTest, BuildpackReference};
+/// use libcnb_test::{IntegrationTest, BuildpackReference, assert_contains};
 ///
 /// # fn call_test_fixture_service(addr: std::net::SocketAddr, payload: &str) -> Result<String, ()> {
 /// #    unimplemented!()
@@ -45,11 +45,11 @@ use std::process::{Command, Stdio};
 ///         BuildpackReference::Crate,
 ///     ])
 ///     .run_test(|context| {
-///         assert!(context.pack_stdout.contains("---> Maven Buildpack"));
-///         assert!(context.pack_stdout.contains("---> Installing Maven"));
-///         assert!(context.pack_stdout.contains("---> Running mvn package"));
+///         assert_contains!(context.pack_stdout, "---> Maven Buildpack");
+///         assert_contains!(context.pack_stdout, "---> Installing Maven");
+///         assert_contains!(context.pack_stdout, "---> Running mvn package");
 ///
-///         context.start_container(&[12345], |container| {
+///         context.prepare_container().expose_port(12345).start_with_default_process(|container| {
 ///             assert_eq!(
 ///                 call_test_fixture_service(
 ///                     container.address_for_port(12345).unwrap(),
@@ -66,6 +66,7 @@ pub struct IntegrationTest {
     target_triple: String,
     builder_name: String,
     buildpacks: Vec<BuildpackReference>,
+    env: HashMap<String, String>,
     docker: Docker,
     tokio_runtime: tokio::runtime::Runtime,
 }
@@ -97,12 +98,12 @@ impl IntegrationTest {
             Ok(docker_host)
                 if docker_host.starts_with("tcp://") || docker_host.starts_with("https://") =>
             {
+                #[cfg(not(feature = "remote-docker"))]
+                panic!("Cannot connect to DOCKER_HOST '{docker_host}' since it requires TLS. Please use a local Docker daemon instead (recommended), or else enable the experimental `remote-docker` feature.");
+                #[cfg(feature = "remote-docker")]
                 Docker::connect_with_ssl_defaults()
             }
-            Ok(docker_host) => panic!(
-                "Cannot connect to unsupported DOCKER_HOST '{}'",
-                docker_host
-            ),
+            Ok(docker_host) => panic!("Cannot connect to unsupported DOCKER_HOST '{docker_host}'"),
             Err(VarError::NotPresent) => Docker::connect_with_local_defaults(),
             Err(VarError::NotUnicode(_)) => {
                 panic!("DOCKER_HOST environment variable is not unicode encoded!")
@@ -115,6 +116,7 @@ impl IntegrationTest {
             target_triple: String::from("x86_64-unknown-linux-musl"),
             builder_name: builder_name.into(),
             buildpacks: vec![BuildpackReference::Crate],
+            env: HashMap::new(),
             docker,
             tokio_runtime,
         }
@@ -136,6 +138,53 @@ impl IntegrationTest {
         self
     }
 
+    /// Inserts or updates an environment variable mapping for the build process.
+    ///
+    /// Note: This does not set this environment variable for running containers, it's only
+    /// available during the build.
+    ///
+    /// # Example
+    /// ```no_run
+    /// use libcnb_test::IntegrationTest;
+    ///
+    /// IntegrationTest::new("heroku/buildpacks:20", "test-fixtures/app")
+    ///     .env("ENV_VAR_ONE", "VALUE ONE")
+    ///     .env("ENV_VAR_TWO", "SOME OTHER VALUE")
+    ///     .run_test(|context| {
+    ///         // ...
+    ///     })
+    /// ```
+    pub fn env(&mut self, k: impl Into<String>, v: impl Into<String>) -> &mut Self {
+        self.env.insert(k.into(), v.into());
+        self
+    }
+
+    /// Adds or updates multiple environment variable mappings for the build process.
+    ///
+    /// Note: This does not set environment variables for running containers, they're only
+    /// available during the build.
+    ///
+    /// # Example
+    /// ```no_run
+    /// use libcnb_test::IntegrationTest;
+    ///
+    /// IntegrationTest::new("heroku/buildpacks:20", "test-fixtures/app")
+    ///     .envs(vec![("ENV_VAR_ONE", "VALUE ONE"), ("ENV_VAR_TWO", "SOME OTHER VALUE")])
+    ///     .run_test(|context| {
+    ///         // ...
+    ///     })
+    /// ```
+    pub fn envs<K: Into<String>, V: Into<String>, I: IntoIterator<Item = (K, V)>>(
+        &mut self,
+        envs: I,
+    ) -> &mut Self {
+        envs.into_iter().for_each(|(key, value)| {
+            self.env(key.into(), value.into());
+        });
+
+        self
+    }
+
     /// Starts a new integration test run.
     ///
     /// This function will copy the application to a temporary directory, cross-compiles this crate,
@@ -153,13 +202,13 @@ impl IntegrationTest {
     ///
     /// # Example
     /// ```no_run
-    /// use libcnb_test::IntegrationTest;
+    /// use libcnb_test::{IntegrationTest, assert_contains};
     ///
     /// IntegrationTest::new("heroku/buildpacks:20", "test-fixtures/app")
     ///     .run_test(|context| {
-    ///         assert!(context.pack_stdout.contains("---> Ruby Buildpack"));
-    ///         assert!(context.pack_stdout.contains("---> Installing bundler"));
-    ///         assert!(context.pack_stdout.contains("---> Installing gems"));
+    ///         assert_contains!(context.pack_stdout, "---> Ruby Buildpack");
+    ///         assert_contains!(context.pack_stdout, "---> Installing bundler");
+    ///         assert_contains!(context.pack_stdout, "---> Installing gems");
     ///     })
     /// ```
     pub fn run_test<F: FnOnce(IntegrationTestContext)>(&mut self, f: F) {
@@ -186,8 +235,17 @@ impl IntegrationTest {
 
         let image_name = util::random_docker_identifier();
 
-        let mut pack_command =
-            PackBuildCommand::new(&self.builder_name, temp_app_dir.path(), &image_name);
+        let mut pack_command = PackBuildCommand::new(
+            &self.builder_name,
+            temp_app_dir.path(),
+            &image_name,
+            // Prevent redundant image-pulling, which slows tests and risks hitting registry rate limits.
+            PullPolicy::IfNotPresent,
+        );
+
+        self.env.iter().for_each(|(key, value)| {
+            pack_command.env(key, value);
+        });
 
         for buildpack in &self.buildpacks {
             match buildpack {
@@ -249,43 +307,27 @@ pub struct IntegrationTestContext<'a> {
 }
 
 impl<'a> IntegrationTestContext<'a> {
-    /// Starts a new container with the image from the integration test.
+    /// Prepares a new container with the image from the integration test.
     ///
-    /// The given `exposed_ports` are mapped to random ports on the host machine. Use
-    /// [`ContainerContext::address_for_port`] to obtain the local port for a mapped port.
+    /// This will not create nor run the container immediately. Use the returned
+    /// `PrepareContainerContext` to configure the container, then call
+    /// [`start_with_default_process`](PrepareContainerContext::start_with_default_process) on it
+    /// to actually create and start the container.
     ///
-    /// # Panics
-    /// - When the container could not be created
-    /// - When the container could not be started
-    pub fn start_container<F: FnOnce(ContainerContext)>(&self, exposed_ports: &[u16], f: F) {
-        let container_name = util::random_docker_identifier();
-
-        self.integration_test.tokio_runtime.block_on(async {
-            self.integration_test
-                .docker
-                .create_container(
-                    Some(CreateContainerOptions {
-                        name: container_name.clone(),
-                    }),
-                    Config {
-                        image: Some(self.image_name.clone()),
-                        ..container_port_mapping::port_mapped_container_config(exposed_ports)
-                    },
-                )
-                .await
-                .expect("Could not create container");
-
-            self.integration_test
-                .docker
-                .start_container(&container_name, None::<StartContainerOptions<String>>)
-                .await
-                .expect("Could not start container");
-        });
-
-        f(ContainerContext {
-            container_name,
-            integration_test_context: self,
-        });
+    /// # Example:
+    ///
+    /// ```no_run
+    /// use libcnb_test::IntegrationTest;
+    ///
+    /// IntegrationTest::new("heroku/buildpacks:20", "test-fixtures/empty-app").run_test(|context| {
+    ///     context.prepare_container().start_with_default_process(|container| {
+    ///         // ...
+    ///     });
+    /// });
+    /// ```
+    #[must_use]
+    pub fn prepare_container(&self) -> PrepareContainerContext {
+        PrepareContainerContext::new(self)
     }
 }
 
